#include <cxxtest/TestSuite.h>

#include "Error.h"
#include "HeapData.h"
#include "MockErrno.h"
#include "MockForRawSocket.h"
#include "MockSelector.h"
#include "SelectableRawSocket.h"
#include "RealMalloc.h"

class DataHandlerForTests : public ISelectableRawSocket::IRawDataHandler
{
public:
    DataHandlerForTests()
    {
        handleDataCalled = false;
    }

    HeapData lastHandledData;
    bool handleDataCalled;
    struct sockaddr_in lastSourceAddress;

    void handleData( const ConstSimpleData &data, struct sockaddr_in &sourceAddress )
    {
        handleDataCalled = true;
        lastHandledData = data;
        memcpy( &lastSourceAddress, &sourceAddress, sizeof(sourceAddress) );
    }
};

class SelectableRawSocketTestSuite : public CxxTest::TestSuite
{
public:
	MockForRawSocket *mock;
	MockSelector *selector;
    DataHandlerForTests *dataHandler;

	void setUp()
	{
		TS_ASSERT( mock = new MockForRawSocket );
		TS_ASSERT( selector = new MockSelector );
        TS_ASSERT( dataHandler = new DataHandlerForTests );
	}

	void tearDown()
	{
        TS_ASSERT_THROWS_NOTHING( delete dataHandler );
		TS_ASSERT_THROWS_NOTHING( delete selector );
		TS_ASSERT_THROWS_NOTHING( delete mock );
	}

	void test_construcor_and_activate()
	{
		SelectableRawSocket *selectableRawSocket;

		TS_ASSERT( selectableRawSocket = new SelectableRawSocket( *selector, "127.0.0.1", *dataHandler ) );

        TS_ASSERT( !mock->socketWasCalled );

        TS_ASSERT_THROWS_NOTHING( selectableRawSocket->activate() );

        TS_ASSERT( mock->socketWasCalled );

        TS_ASSERT( selector->registerWasCalled );
		TS_ASSERT_EQUALS( selector->lastRegistetedDescriptor, mock->nextSocketDescriptor );
        TS_ASSERT( !selector->unregisterWasCalled );

		TS_ASSERT( mock->bindWasCalled );

        sockaddr_in expectedBindSockaddr;
		memset( &expectedBindSockaddr, 0, sizeof(sockaddr_in ) );

		expectedBindSockaddr.sin_family = AF_INET;
        inet_aton( "127.0.0.1", &expectedBindSockaddr.sin_addr );

        TS_ASSERT_SAME_DATA( &mock->lastBindSockaddr, &expectedBindSockaddr, sizeof(sockaddr_in) );

        TS_ASSERT_THROWS_NOTHING( delete selectableRawSocket );

		TS_ASSERT( mock->closeWasCalled );

		TS_ASSERT( selector->unregisterWasCalled );
		TS_ASSERT_EQUALS( selector->lastUnregistetedDescriptor, mock->nextSocketDescriptor );
	}

	void test_constructor_and_activate__socket_fails()
	{
        SelectableRawSocket *selectableRawSocket;

		mock->nextSocketDescriptor = -1;

        TS_ASSERT( selectableRawSocket = new SelectableRawSocket( *selector, "127.0.0.1", *dataHandler ) );

        TS_ASSERT_THROWS_EQUALS( selectableRawSocket->activate(),
								 const Error &e,
								 e.code(),
								 Error::SOCKET_CREATION_FAILED );

		TS_ASSERT( !selector->registerWasCalled );
		TS_ASSERT( !mock->closeWasCalled );
	}

	void test_constructor__bad_address()
	{
        SelectableRawSocket *selectableRawSocket;

        mock->nextBindResult = -1;

        TS_ASSERT( selectableRawSocket = new SelectableRawSocket( *selector, "127.0.0.1.bla", *dataHandler ) );

        TS_ASSERT_THROWS_EQUALS( selectableRawSocket->activate(),
                                 const Error &e,
								 e.code(),
								 Error::INET_ATON_FAILED );

		TS_ASSERT( !selector->registerWasCalled );
		TS_ASSERT( !selector->unregisterWasCalled );
		TS_ASSERT( mock->closeWasCalled );
	}

	void test_constructor__bind_fails()
	{
        SelectableRawSocket *selectableRawSocket;

        mock->nextBindResult = -1;

        TS_ASSERT( selectableRawSocket = new SelectableRawSocket( *selector, "127.0.0.1", *dataHandler ) );

        TS_ASSERT_THROWS_EQUALS( selectableRawSocket->activate(),
                                 const Error &e,
								 e.code(),
								 Error::BIND_FAILED );

		TS_ASSERT( !selector->registerWasCalled );
		TS_ASSERT( !selector->unregisterWasCalled );
		TS_ASSERT( mock->closeWasCalled );
	}

	void test_close()
	{
		SelectableRawSocket *selectableRawSocket;

        TS_ASSERT( selectableRawSocket = new SelectableRawSocket( *selector, "127.0.0.1", *dataHandler ) );
        TS_ASSERT_THROWS_NOTHING( selectableRawSocket->activate() );

		TS_ASSERT( !mock->closeWasCalled );
        TS_ASSERT( !selector->unregisterWasCalled );

        TS_ASSERT_THROWS_NOTHING( selectableRawSocket->close() );

		TS_ASSERT( mock->closeWasCalled );
        TS_ASSERT( selector->unregisterWasCalled );
		TS_ASSERT_EQUALS( selector->lastUnregistetedDescriptor, mock->nextSocketDescriptor );

        mock->closeWasCalled = false;

        TS_ASSERT_THROWS_NOTHING( delete selectableRawSocket );

		TS_ASSERT( !mock->closeWasCalled );
	}

    void test_can_read()
    {
		SelectableRawSocket *selectableRawSocket;

        TS_ASSERT( selectableRawSocket = new SelectableRawSocket( *selector, "127.0.0.1", *dataHandler ) );

        TS_ASSERT_THROWS_NOTHING( selectableRawSocket->activate() );

        char buffer[] = { 0x0a, 0x0b, 0x0c, 0x0d, 0x0e };

        mock->nextReceiveData = ConstSimpleData( buffer, sizeof(buffer) );

        inet_aton( "82.32.11.2", &mock->nextSourceAddress.sin_addr );
        mock->nextSourceAddress.sin_port = 5588;

        TS_ASSERT_THROWS_NOTHING( selectableRawSocket->canRead() );

        TS_ASSERT_EQUALS( dataHandler->lastHandledData.size(), sizeof(buffer) );
        TS_ASSERT_SAME_DATA( dataHandler->lastHandledData.data(), buffer, sizeof(buffer) );

        TS_ASSERT_SAME_DATA( &mock->nextSourceAddress, &dataHandler->lastSourceAddress, sizeof(struct sockaddr_in) );

        TS_ASSERT_THROWS_NOTHING( delete selectableRawSocket );
    }

    void test_can_read__fails()
    {
		SelectableRawSocket *selectableRawSocket;

        TS_ASSERT( selectableRawSocket = new SelectableRawSocket( *selector, "127.0.0.1", *dataHandler ) );

        TS_ASSERT_THROWS_NOTHING( selectableRawSocket->activate() );

        char buffer[] = { 0x0a, 0x0b, 0x0c, 0x0d, 0x0e };

        mock->nextReceiveData = ConstSimpleData( buffer, sizeof(buffer) );

        inet_aton( "82.32.11.2", &mock->nextSourceAddress.sin_addr );
        mock->nextSourceAddress.sin_port = 5588;

        mock->recvShouldFail = true;

        TS_ASSERT_THROWS_EQUALS( selectableRawSocket->canRead(),
                                 const Error &e,
                                 e.code(),
                                 Error::RECV_FAILED );

        TS_ASSERT_THROWS_NOTHING( delete selectableRawSocket );
    }

    void test_send__fails()
    {
		SelectableRawSocket *selectableRawSocket;

        TS_ASSERT( selectableRawSocket = new SelectableRawSocket( *selector, "127.0.0.1", *dataHandler ) );

        TS_ASSERT_THROWS_NOTHING( selectableRawSocket->activate() );

        mock->sendShouldFail = true;

        char buffer[] = { 0x0a, 0x0b, 0x0c, 0x0d, 0x0e };

        TS_ASSERT_THROWS_EQUALS( selectableRawSocket->send( ConstSimpleData( buffer, sizeof(buffer) ), "182.32.11.22" ),
                                 const Error &e,
                                 e.code(),
                                 Error::SEND_FAILED );

        TS_ASSERT_THROWS_NOTHING( delete selectableRawSocket );

    }

    void test_send()
    {
		SelectableRawSocket *selectableRawSocket;

        TS_ASSERT( selectableRawSocket = new SelectableRawSocket( *selector, "127.0.0.1", *dataHandler ) );

        TS_ASSERT_THROWS_NOTHING( selectableRawSocket->activate() );

        char buffer[] = { 0x0a, 0x0b, 0x0c, 0x0d, 0x0e };

        TS_ASSERT_THROWS_NOTHING( selectableRawSocket->send( ConstSimpleData( buffer, sizeof(buffer) ), "182.32.11.22" ) );

        TS_ASSERT_EQUALS( mock->lastSentData.size(), sizeof(buffer) );
        TS_ASSERT_SAME_DATA( mock->lastSentData.data(), buffer, sizeof(buffer) );

        struct sockaddr_in expectedAddress;

        memset( &expectedAddress, 0, sizeof(expectedAddress) );
        inet_aton( "182.32.11.22", &expectedAddress.sin_addr );

        TS_ASSERT_SAME_DATA( &expectedAddress, &mock->lastSendAddress, sizeof(struct sockaddr_in) );

        TS_ASSERT_THROWS_NOTHING( delete selectableRawSocket );
    }
};

//
// Local Variables:
// compile-command: "make -C ../.. "
// tags-file-name: "../../TAGS"
// c-basic-offset: 4
// End:
//
