#ifndef __BThread_h__
#define __BThread_h__

#include "AutoSelectableSocket.h"
#include "AutoSelector.h"
#include "AutoSocket.h"
#include "AutoSyncPointResolver.h"
#include "EngineConfiguration.h"
#include "Event.h"
#include "EventPredicate.h"
#include "IBThread.h"
#include "ISelectableSocket.h"
#include "SyncRequest.h"
#include "Vector.h"
#include <arpa/inet.h>

class BThread : public IBThread
{
public:
    enum ThreadType {
        REGULAR_THREAD,
        NON_BLOCKING_THREAD,
    };

    BThread( ThreadType threadType ) : _socket( inet_addr( EngineConfiguration::LOOPBACK_IP ), enginePort( threadType ) ),
                                       _syncPointResolver( _selector, _socket ),
                                       _selectableSocket( _socket, _selector, _syncPointResolver )
    {
    }

    IBThread::SyncResult bSync( const Vector<Event> &requested,
                                const Vector<Event> &watched,
                                const Vector<Event> &blocked,
                                unsigned timeoutInSeconds,
                                String label = "" )
    {
        AtomicPropositions noPropositions;

        _syncPointResolver->go( requested, watched, blocked, timeoutInSeconds, noPropositions, label );
        return selectUntilResult();
    }

    IBThread::SyncResult bSync( const Vector<Event> &requested,
                                const Vector<Event> &watched,
                                const Vector<Event> &blocked,
                                unsigned timeoutInSeconds,
                                const AtomicPropositions &atomicPropositions,
                                String label = "" )
    {
        _syncPointResolver->go( requested, watched, blocked, timeoutInSeconds, atomicPropositions, label );
        return selectUntilResult();
    }

    IBThread::SyncResult bSync( const Vector<Event> &requested,
                                const Vector<Event> &watched,
                                EventPredicate *blockPredicate,
                                unsigned timeoutInSeconds,
                                String label = "" )
    {
        AtomicPropositions noPropositions;

        _syncPointResolver->go( requested, watched, blockPredicate, timeoutInSeconds, noPropositions, label );
        return selectUntilResult();
    }

    IBThread::SyncResult bSync( const Vector<Event> &requested,
                                const Vector<Event> &watched,
                                EventPredicate *blockPredicate,
                                unsigned timeoutInSeconds,
                                const AtomicPropositions &atomicPropositions,
                                String label = "" )
    {
        _syncPointResolver->go( requested, watched, blockPredicate, timeoutInSeconds, atomicPropositions, label );
        return selectUntilResult();
    }

    Event lastEvent() const
    {
        return _lastEvent;
        //return _syncPointResolver->lastEvent();
    }

    virtual void done()
    {
    }

    unsigned enginePort( ThreadType threadType )
    {
        return ( threadType == REGULAR_THREAD ) ? EngineConfiguration::BTHREAD_MANAGEMENT_PORT : EngineConfiguration::NON_BLOCKING_BTHREAD_MANAGEMENT_PORT;
    }

    void sleepForever()
    {
        clearEventVectors();
        AtomicPropositions noPropositions;

        _syncPointResolver->go( _requested, _watched, _blocked, 0, noPropositions, "Asleep" );
        selectUntilResult();
    }

protected:
    Vector<Event> _requested;
    Vector<Event> _watched;
    Vector<Event> _blocked;

    void clearEventVectors()
    {
        _requested.clear();
        _watched.clear();
        _blocked.clear();
    }

private:
    AutoSelector _selector;
    AutoSocket _socket;
    Event _lastEvent;
    AutoSyncPointResolver _syncPointResolver;
    AutoSelectableSocket _selectableSocket;

    IBThread::SyncResult selectUntilResult()
    {
        while ( !_syncPointResolver->done() )
        {
            _selector->select();
        }

        _lastEvent = _syncPointResolver->lastEvent();
        return _lastEvent.type() == EVENT_TYPE_IF_EXPIRED ? TIMEOUT : EVENT_WAS_TRIGGERED;

        //return _syncPointResolver->lastEventIsTimeout() ? IBThread::TIMEOUT : IBThread::EVENT_WAS_TRIGGERED;
    }
};

#endif // __BThread_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
