#ifndef __SyncPointResolver_h__
#define __SyncPointResolver_h__

#include "AtomicPropositions.h"
#include "AutoTimer.h"
#include "Event.h"
#include "IBThread.h"
#include "ISelectableSocket.h"
#include "ISyncPointResolver.h"
#include "SyncRequest.h"
#include "Vector.h"

class SyncPointResolver : public ITimer::IExpired,  public ISyncPointResolver
{
public:
    SyncPointResolver( ISelector &selector, ISocket &socket ) : _socket( socket ), _timer( selector, *this ), _done( false )
    {
    }

    void go( const Vector<Event> &requested,
             const Vector<Event> &watched,
             const Vector<Event> &blocked,
             unsigned timeoutInSeconds,
             const AtomicPropositions &,
             const String &label )
    {
        storeEventsAndLabel( requested, watched, blocked, timeoutInSeconds, label );
        resolveSyncPoint();
    }

    void go( const Vector<Event> &requested,
             const Vector<Event> &watched,
             EventPredicate *blockPredicate,
             unsigned timeoutInSeconds,
             const AtomicPropositions &,
             const String &label )
    {
        storeEventsAndLabel( requested, watched, blockPredicate, timeoutInSeconds, label );
        resolveSyncPoint();
    }

    bool done() const
    {
        return _done;
    }

    void handleData( const ConstSimpleData &data )
    {
        _lastEvent.deserialize( data );

        if ( relevantEventTriggered() )
        {
            _timer->stop();
            _done = true;
        }
        else
        {
            sendSyncRequest();
        }
    }

    Event lastEvent() const
    {
        return _lastEvent;
    }

    void expired()
    {
        _done = true;
        _lastEvent = Event( IBThread::EVENT_TYPE_IF_EXPIRED );
    }

    bool lastEventIsTimeout() const
    {
        return _lastEvent.type() == IBThread::EVENT_TYPE_IF_EXPIRED;
    }

private:
    ISocket &_socket;
    Event _lastEvent;
    AutoTimer _timer;
    bool _done;

    Vector<Event> _requested;
    Vector<Event> _watched;
    Vector<Event> _blocked;
    EventPredicate *_blockPredicate;
    String _label;
    unsigned _timeoutInSeconds;

    bool _usingBlockedList;

    void sendSyncRequest()
    {
        AtomicPropositions noPropositions;
        HeapData serialized;

        if ( _usingBlockedList )
        {
            SyncRequest( _requested, _blocked, noPropositions, _label ).serialize( serialized );
        }
        else
        {
            SyncRequest( _requested, _blockPredicate, noPropositions, _label ).serialize( serialized );
        }

        _socket.send( serialized );
    }

    bool relevantEventTriggered()
    {
        if ( _watched.exists( _lastEvent ) )
        {
            return true;
        }

        Vector<Event>::iterator it;

        for ( it = _requested.begin(); it != _requested.end(); ++it )
        {
            if ( it->type() == _lastEvent.type() )
            {
                if ( it->payloadSize() == _lastEvent.payloadSize() )
                {
                    if ( it->payloadSize() == 0 )
                    {
                        return true;
                    }
                    else if ( memcmp( it->payload().data(), _lastEvent.payload().data(), it->payloadSize() ) == 0 )
                    {
                        return true;
                    }
                }
            }
        }

        return false;
    }

    void resetDone()
    {
        _done = false;
    }

    void storeEventsAndLabel( const Vector<Event> &requested, const Vector<Event> &watched, const Vector<Event> &blocked, unsigned timeoutInSeconds, const String &label )
    {
        _requested = requested;
        _watched = watched;
        _blocked = blocked;
        _usingBlockedList = true;
        _label = label;
        _timeoutInSeconds = timeoutInSeconds;
    }

    void storeEventsAndLabel( const Vector<Event> &requested, const Vector<Event> &watched, EventPredicate *blockPredicate, unsigned timeoutInSeconds, const String &label )
    {
        _requested = requested;
        _watched = watched;
        _blockPredicate = blockPredicate;
        _usingBlockedList = false;
        _label = label;
        _timeoutInSeconds = timeoutInSeconds;
    }

    void resolveSyncPoint()
    {
        resetDone();
        sendSyncRequest();

        if ( _timeoutInSeconds > 0 )
        {
            _timer->set( _timeoutInSeconds );
        }
    }
};

#endif // __SyncPointResolver_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
