#ifndef __SyncRequest_h__
#define __SyncRequest_h__

#include "AtomicPropositions.h"
#include "Event.h"
#include "EventPredicate.h"
#include "Vector.h"

class ListOfEvents
{
public:
    ListOfEvents()
    {
    }

    ListOfEvents( const Vector<Event> &events ) : _events( events )
    {
    }

    void serialize( HeapData &serialized )
    {
        addNumberOfEvents( serialized );
        for ( unsigned i = 0; i < _events.size(); ++i )
        {
            addEvent( serialized, i );
        }
    }

    void deserialize( const ConstSimpleData &serialized )
    {
        const char *asChar = serialized.asChar();

        unsigned eventCount = extractUnsigned( asChar );
        asChar += sizeof(unsigned);

        for ( unsigned i = 0; i < eventCount; ++i )
        {
            unsigned singleEventSize = extractUnsigned( asChar );
            asChar += sizeof(unsigned);

            createNewEvent( asChar, singleEventSize );
            asChar += singleEventSize;
        }

        _totalSize = asChar - ( (char*)serialized.data() );

        _createdEvents = true;
    }

    unsigned totalSize()
    {
        return _totalSize;
    }

    Vector<Event> _events;
private:
    unsigned _totalSize;
    bool _createdEvents;

    void addNumberOfEvents( HeapData &serialized )
    {
        unsigned numberOfEvents = _events.size();
        serialized = ConstSimpleData( &numberOfEvents, sizeof(numberOfEvents) );
    }

    void addEvent( HeapData &serialized, unsigned index )
    {
        HeapData serializedEvent;
        _events[index].serialize( serializedEvent );
        unsigned eventSize = serializedEvent.size();

        serialized += ConstSimpleData( &eventSize, sizeof(eventSize) );
        serialized += serializedEvent;
    }

    void createNewEvent( const char *serialized, unsigned size )
    {
        Event event;
        event.deserialize( ConstSimpleData( serialized, size ) );
        _events.append( event );
    }

    unsigned extractUnsigned( const char *serialized )
    {
        return *(unsigned *)serialized;
    }
};

class SyncRequest
{
public:
    enum TypeOfBlocked {
        BLOCKED_LIST = 0,
        BLOCKED_PREDICATE = 1,
    };

    SyncRequest( const Vector<Event> &requested, const Vector<Event> &blocked, const AtomicPropositions &atomicPropositions, String label = "" ) :
        _requested( requested ),
        _blocked( blocked ),
        _atomicPropositions( atomicPropositions ),
        _label( label ),
        _typeOfBlocked( BLOCKED_LIST )
    {
    }

    SyncRequest( const Vector<Event> &requested, EventPredicate *blockPredicate, const AtomicPropositions &atomicPropositions, String label = "" ) :
        _requested( requested ),
        _blockPredicate( blockPredicate ),
        _atomicPropositions( atomicPropositions ),
        _label( label ),
        _typeOfBlocked( BLOCKED_PREDICATE )
    {
    }

    void serialize( HeapData &serialized )
    {
        addLabel( serialized );
        addRequested( serialized );
        addBlocked( serialized );
        addAtomicPropositions( serialized );
    }

    SyncRequest( const ConstSimpleData &serialized )
    {
        unsigned offset;

        extractLabel( serialized, offset );
        extractRequested( serialized, offset );
        extractBlocked( serialized, offset );
        extractAtomicPropositions( serialized, offset );
    }

    Vector<Event> getRequested()
    {
        return _requested._events;
    }

    TypeOfBlocked typeOfBlocked() const
    {
        return (TypeOfBlocked)_typeOfBlocked;
    }

    Vector<Event> getBlockedEvents()
    {
        if ( _typeOfBlocked != BLOCKED_LIST )
        {
            throw Error( Error::WRONG_TYPE_OF_BLOCKED_EVENTS );
        }

        return _blocked._events;
    }

    EventPredicate *getBlockedPredicate()
    {
        if ( _typeOfBlocked != BLOCKED_PREDICATE )
        {
            throw Error( Error::WRONG_TYPE_OF_BLOCKED_EVENTS );
        }

        return _blockPredicate;
    }

    String getLabel()
    {
        return _label;
    }

    AtomicPropositions getAtomicPropositions()
    {
        return _atomicPropositions;
    }

private:
    ListOfEvents _requested;
    ListOfEvents _blocked;
    EventPredicate *_blockPredicate;
    AtomicPropositions _atomicPropositions;
    String _label;
    unsigned _typeOfBlocked;

    void addLabel( HeapData &serialized )
    {
        unsigned labelLength = _label.length();
        serialized = ConstSimpleData( &labelLength, sizeof(labelLength) );

        ConstSimpleData labelString( _label.ascii(), _label.length() );
        serialized += labelString;
    }

    void addAtomicPropositions( HeapData &serialized )
    {
        unsigned numberOfPropositions = _atomicPropositions.size();
        serialized += ConstSimpleData( &numberOfPropositions, sizeof(numberOfPropositions) );

        for ( unsigned i = 0; i < numberOfPropositions; ++i )
        {
            String proposition = _atomicPropositions.at( i );
            unsigned propositionLength = proposition.length();
            serialized += ConstSimpleData( &propositionLength, sizeof(propositionLength) );

            ConstSimpleData propositionString( proposition.ascii(), propositionLength );
            serialized += propositionString;
        }
    }

    void addRequested( HeapData &serialized )
    {
        HeapData temp;
        _requested.serialize( temp );
        serialized += temp;
    }

    void addBlocked( HeapData &serialized )
    {
        serialized += ConstSimpleData( &_typeOfBlocked, sizeof(_typeOfBlocked) );

        if ( _typeOfBlocked == BLOCKED_LIST )
        {
            HeapData temp;
            _blocked.serialize( temp );
            serialized += temp;
        }
        else
        {
            serialized += ConstSimpleData( &_blockPredicate, sizeof(_blockPredicate) );
        }
    }

    void extractLabel( const ConstSimpleData &serialized, unsigned &offset )
    {
        unsigned labelLength;
        memcpy( &labelLength, serialized.asChar(), sizeof(labelLength) );

        offset = sizeof(unsigned);

        _label = String( serialized.asChar() + offset, labelLength );

        offset += labelLength;
    }

    void extractRequested( const ConstSimpleData &serialized, unsigned &offset )
    {
        _requested.deserialize( ConstSimpleData( serialized.asChar() + offset, serialized.size() - offset ) );
        offset += _requested.totalSize();
    }

    void extractBlocked( const ConstSimpleData &serialized, unsigned &offset )
    {
        extractBlockedType( serialized, offset );

        if ( _typeOfBlocked == BLOCKED_LIST )
        {
            _blocked.deserialize( ConstSimpleData( serialized.asChar() + offset, serialized.size() - offset ) );
            offset += _blocked.totalSize();
        }
        else
        {
            memcpy( &_blockPredicate, serialized.asChar() + offset, sizeof(_blockPredicate) );
            offset += sizeof(_blockPredicate);
        }
    }

    void extractBlockedType( const ConstSimpleData &serialized, unsigned &offset )
    {
        memcpy( &_typeOfBlocked, serialized.asChar() + offset, sizeof(_typeOfBlocked) );
        offset += sizeof(_typeOfBlocked );
    }

    void extractAtomicPropositions( const ConstSimpleData &serialized, unsigned &offset )
    {
        unsigned numberOfPropositions;
        memcpy( &numberOfPropositions, serialized.asChar() + offset, sizeof(numberOfPropositions) );
        offset += sizeof(numberOfPropositions);

        for ( unsigned i = 0; i < numberOfPropositions; ++i )
        {
            unsigned propositionLength;
            memcpy( &propositionLength, serialized.asChar() + offset, sizeof(propositionLength) );
            offset += sizeof(propositionLength);

            _atomicPropositions.add( String( serialized.asChar() + offset, propositionLength ) );
            offset += propositionLength;
        }
    }
};

#endif // __SyncRequest_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
