#include <cxxtest/TestSuite.h>

#include "BlockPredicateForTests.h"
#include "Event.h"
#include "IBThread.h"
#include "MockSelectableSocketFactory.h"
#include "MockSelectorFactory.h"
#include "MockSocketFactory.h"
#include "MockTimerFactory.h"
#include "RealMalloc.h"
#include "SyncPointResolver.h"

class MockForSyncPointResolver :
    public RealMalloc,
    public MockSocketFactory,
    public MockTimerFactory,
    public MockSelectorFactory,
    public MockSelectableSocketFactory
{
public:
};

class SyncPointResolverTestSuite : public CxxTest::TestSuite
{
public:
	MockForSyncPointResolver *mock;
    MockSocket *socket;
    MockTimer *timer;
    MockSelector *selector;
    AtomicPropositions noPropositions;

	void setUp()
	{
		TS_ASSERT( mock = new MockForSyncPointResolver );
        TS_ASSERT( selector = new MockSelector );
        TS_ASSERT( socket = new MockSocket );

        timer = &( mock->mockTimer );
	}

	void tearDown()
	{
		TS_ASSERT_THROWS_NOTHING( delete socket );
		TS_ASSERT_THROWS_NOTHING( delete selector );
		TS_ASSERT_THROWS_NOTHING( delete mock );
	}

    void test_constructor_destructor()
    {
        SyncPointResolver *syncPointResolver;

        TS_ASSERT( syncPointResolver = new SyncPointResolver( *selector, *socket ) );

        TS_ASSERT( timer->wasCreated );
        TS_ASSERT_EQUALS( timer->lastExpired, syncPointResolver );
        TS_ASSERT_EQUALS( timer->lastSelector, selector );

        TS_ASSERT_THROWS_NOTHING( delete syncPointResolver );

        TS_ASSERT( timer->wasDiscarded );
    }

    void test_go__no_timeout()
    {
        Event event1( 1 );

        char buffer[] = { 0x01, 0x02 };
        Event event2( 2, ConstSimpleData( buffer, sizeof(buffer) ) );

        Event event3( 3 );
        Event event4( 4 );

        SyncPointResolver syncPointResolver( *selector, *socket );

        Vector<Event> requested;
        requested.append( event1 );
        requested.append( event2 );

        Vector<Event> watched;
        watched.append( event3 );

        Vector<Event> blocked;
        blocked.append( event4 );

        char socketData1[] = { 0x01, 0x02, 0x03, 0x04, 0x00, 0x00, 0x00, 0x00 };
        ConstSimpleData constData1( socketData1, sizeof(socketData1) );

        char socketData2[] = { 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
        ConstSimpleData constData2( socketData2, sizeof(socketData2) );

        String label = "hi";

        TS_ASSERT( !syncPointResolver.done() );

        TS_ASSERT_THROWS_NOTHING( syncPointResolver.go( requested, watched, blocked, IBThread::NO_TIMEOUT, noPropositions, label ) );

        TS_ASSERT( !syncPointResolver.done() );

        char expectedData[] = {
            // Label
            0x02, 0x00, 0x00, 0x00,
            0x68, 0x69,

            // Requested
            0x02, 0x00, 0x00, 0x00, // Number of events
            0x08, 0x00, 0x00, 0x00, // Length of first
            0x01, 0x00, 0x00, 0x00, // First
            0x00, 0x00, 0x00, 0x00,
            0x0a, 0x00, 0x00, 0x00, // Length of Second
            0x02, 0x00, 0x00, 0x00, // Second
            0x02, 0x00, 0x00, 0x00,
            0x01, 0x02,

            // Blocked
            0x00, 0x00, 0x00, 0x00, // Type = List
            0x01, 0x00, 0x00, 0x00, // Number of events
            0x08, 0x00, 0x00, 0x00, // Length of first
            0x04, 0x00, 0x00, 0x00, //
            0x00, 0x00, 0x00, 0x00,

            // Atomic Propositions
            0x00, 0x00, 0x00, 0x00 // Number of propositions
        };

        TS_ASSERT_EQUALS( socket->lastSentData.size(), 1U );

        TS_ASSERT_EQUALS( socket->lastSentData[0].size(), sizeof(expectedData) );
        TS_ASSERT_SAME_DATA( socket->lastSentData[0].data(), expectedData, sizeof(expectedData) );

        TS_ASSERT( !syncPointResolver.done() );

        TS_ASSERT_THROWS_NOTHING( syncPointResolver.handleData( constData1 ) );

        TS_ASSERT_EQUALS( socket->lastSentData.size(), 2U );

        TS_ASSERT_EQUALS( socket->lastSentData[1].size(), sizeof(expectedData) );
        TS_ASSERT_SAME_DATA( socket->lastSentData[1].data(), expectedData, sizeof(expectedData) );

        TS_ASSERT( !syncPointResolver.done() );

        TS_ASSERT_THROWS_NOTHING( syncPointResolver.handleData( constData2 ) );

        TS_ASSERT( syncPointResolver.done() );
        TS_ASSERT_EQUALS( socket->lastSentData.size(), 2U );

        TS_ASSERT_EQUALS( syncPointResolver.lastEvent().type(), 0x00000001U );
        TS_ASSERT_EQUALS( syncPointResolver.lastEvent().payloadSize(), 0U );
    }

    void test_go__timeout_set__timer_expires()
    {
        Event event1( 1 );

        char buffer[] = { 0x01, 0x02 };
        Event event2( 2, ConstSimpleData( buffer, sizeof(buffer) ) );

        Event event3( 3 );
        Event event4( 4 );

        SyncPointResolver syncPointResolver( *selector, *socket );

        Vector<Event> requested;
        requested.append( event1 );
        requested.append( event2 );

        Vector<Event> watched;
        watched.append( event3 );

        Vector<Event> blocked;
        blocked.append( event4 );

        char socketData1[] = { 0x01, 0x02, 0x03, 0x04, 0x00, 0x00, 0x00, 0x00 };
        ConstSimpleData constData1( socketData1, sizeof(socketData1) );

        String label = "hi";

        TS_ASSERT_THROWS_NOTHING( syncPointResolver.go( requested, watched, blocked, 5, noPropositions, label ) );

        char expectedData[] = {
            // Label
            0x02, 0x00, 0x00, 0x00,
            0x68, 0x69,

            // Requested
            0x02, 0x00, 0x00, 0x00, // Number of events
            0x08, 0x00, 0x00, 0x00, // Length of first
            0x01, 0x00, 0x00, 0x00, // First
            0x00, 0x00, 0x00, 0x00,
            0x0a, 0x00, 0x00, 0x00, // Length of Second
            0x02, 0x00, 0x00, 0x00, // Second
            0x02, 0x00, 0x00, 0x00,
            0x01, 0x02,

            // Blocked
            0x00, 0x00, 0x00, 0x00, // Type = List
            0x01, 0x00, 0x00, 0x00, // Number of events
            0x08, 0x00, 0x00, 0x00, // Length of first
            0x04, 0x00, 0x00, 0x00, //
            0x00, 0x00, 0x00, 0x00,

            // Atomic Propositions
            0x00, 0x00, 0x00, 0x00 // Number of propositions
        };

        TS_ASSERT_EQUALS( socket->lastSentData.size(), 1U );

        TS_ASSERT_EQUALS( socket->lastSentData[0].size(), sizeof(expectedData) );
        TS_ASSERT_SAME_DATA( socket->lastSentData[0].data(), expectedData, sizeof(expectedData) );

        TS_ASSERT( !syncPointResolver.done() );

        TS_ASSERT( timer->setWasCalled );
        TS_ASSERT_EQUALS( timer->lastSetSeconds, 5U );
        timer->setWasCalled = false;

        TS_ASSERT_THROWS_NOTHING( syncPointResolver.handleData( constData1 ) );

        TS_ASSERT( !timer->setWasCalled );

        TS_ASSERT_EQUALS( socket->lastSentData.size(), 2U );

        TS_ASSERT_EQUALS( socket->lastSentData[1].size(), sizeof(expectedData) );
        TS_ASSERT_SAME_DATA( socket->lastSentData[1].data(), expectedData, sizeof(expectedData) );

        TS_ASSERT( !syncPointResolver.done() );

        TS_ASSERT_THROWS_NOTHING( syncPointResolver.expired() );

        TS_ASSERT( syncPointResolver.done() );
        TS_ASSERT_EQUALS( socket->lastSentData.size(), 2U );

        TS_ASSERT_EQUALS( syncPointResolver.lastEvent().type(), IBThread::EVENT_TYPE_IF_EXPIRED );
        TS_ASSERT_EQUALS( syncPointResolver.lastEvent().payloadSize(), 0U );
        TS_ASSERT( syncPointResolver.lastEventIsTimeout() );
    }

    void test_go__timeout_set__timer_doesnt_expire()
    {
        Event event1( 1 );

        char buffer[] = { 0x01, 0x02 };
        Event event2( 2, ConstSimpleData( buffer, sizeof(buffer) ) );

        Event event3( 3 );
        Event event4( 4 );

        SyncPointResolver syncPointResolver( *selector, *socket );

        Vector<Event> requested;
        requested.append( event1 );
        requested.append( event2 );

        Vector<Event> watched;
        watched.append( event3 );

        Vector<Event> blocked;
        blocked.append( event4 );

        char socketData1[] = { 0x01, 0x02, 0x03, 0x04, 0x00, 0x00, 0x00, 0x00 };
        ConstSimpleData constData1( socketData1, sizeof(socketData1) );

        char socketData2[] = { 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
        ConstSimpleData constData2( socketData2, sizeof(socketData2) );

        String label = "hi";

        TS_ASSERT_THROWS_NOTHING( syncPointResolver.go( requested, watched, blocked, 5, noPropositions, label ) );

        char expectedData[] = {
            // Label
            0x02, 0x00, 0x00, 0x00,
            0x68, 0x69,

            // Requested
            0x02, 0x00, 0x00, 0x00, // Number of events
            0x08, 0x00, 0x00, 0x00, // Length of first
            0x01, 0x00, 0x00, 0x00, // First
            0x00, 0x00, 0x00, 0x00,
            0x0a, 0x00, 0x00, 0x00, // Length of Second
            0x02, 0x00, 0x00, 0x00, // Second
            0x02, 0x00, 0x00, 0x00,
            0x01, 0x02,

            // Blocked
            0x00, 0x00, 0x00, 0x00, // Type = List
            0x01, 0x00, 0x00, 0x00, // Number of events
            0x08, 0x00, 0x00, 0x00, // Length of first
            0x04, 0x00, 0x00, 0x00, //
            0x00, 0x00, 0x00, 0x00,

            // Atomic Propositions
            0x00, 0x00, 0x00, 0x00 // Number of propositions
        };

        TS_ASSERT_EQUALS( socket->lastSentData.size(), 1U );

        TS_ASSERT_EQUALS( socket->lastSentData[0].size(), sizeof(expectedData) );
        TS_ASSERT_SAME_DATA( socket->lastSentData[0].data(), expectedData, sizeof(expectedData) );

        TS_ASSERT( !syncPointResolver.done() );

        TS_ASSERT( timer->setWasCalled );
        TS_ASSERT_EQUALS( timer->lastSetSeconds, 5U );
        timer->setWasCalled = false;

        TS_ASSERT_THROWS_NOTHING( syncPointResolver.handleData( constData1 ) );

        TS_ASSERT( !timer->setWasCalled );

        TS_ASSERT_EQUALS( socket->lastSentData.size(), 2U );

        TS_ASSERT_EQUALS( socket->lastSentData[1].size(), sizeof(expectedData) );
        TS_ASSERT_SAME_DATA( socket->lastSentData[1].data(), expectedData, sizeof(expectedData) );

        TS_ASSERT( !syncPointResolver.done() );

        TS_ASSERT_THROWS_NOTHING( syncPointResolver.handleData( constData2 ) );

        TS_ASSERT( timer->stopWasCalled );

        TS_ASSERT( syncPointResolver.done() );
        TS_ASSERT_EQUALS( socket->lastSentData.size(), 2U );

        TS_ASSERT_EQUALS( syncPointResolver.lastEvent().type(), 0x00000001U );
        TS_ASSERT( !syncPointResolver.lastEventIsTimeout() );
        TS_ASSERT_EQUALS( syncPointResolver.lastEvent().payloadSize(), 0U );
    }

    void test_payload_matters_on_requested()
    {
        Event event1( 1 );

        char buffer[] = { 0x01, 0x02 };
        Event event2( 2, ConstSimpleData( buffer, sizeof(buffer) ) );

        Event event3( 3 );
        Event event4( 4 );

        SyncPointResolver syncPointResolver( *selector, *socket );

        Vector<Event> requested;
        requested.append( event1 );
        requested.append( event2 );

        Vector<Event> watched;
        watched.append( event3 );

        Vector<Event> blocked;
        blocked.append( event4 );

        char socketData1[] = { 0x02, 0x00, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0x01, 0x01 };
        ConstSimpleData constData1( socketData1, sizeof(socketData1) );

        char socketData2[] = { 0x02, 0x00, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0x01, 0x02 };
        ConstSimpleData constData2( socketData2, sizeof(socketData2) );

        String label = "hi";

        TS_ASSERT( !syncPointResolver.done() );

        TS_ASSERT_THROWS_NOTHING( syncPointResolver.go( requested, watched, blocked, IBThread::NO_TIMEOUT, noPropositions, label ) );

        TS_ASSERT( !syncPointResolver.done() );

        TS_ASSERT_THROWS_NOTHING( syncPointResolver.handleData( constData1 ) );

        TS_ASSERT( !syncPointResolver.done() );

        TS_ASSERT_THROWS_NOTHING( syncPointResolver.handleData( constData2 ) );

        TS_ASSERT( syncPointResolver.done() );
        TS_ASSERT_EQUALS( socket->lastSentData.size(), 2U );

        TS_ASSERT_EQUALS( syncPointResolver.lastEvent().type(), 0x00000002U );
        TS_ASSERT( !syncPointResolver.lastEventIsTimeout() );
        TS_ASSERT_EQUALS( syncPointResolver.lastEvent().payloadSize(), 2U );
    }

    void test_payload_does_not_matter_on_watched()
    {
        Event event1( 1 );

        char buffer[] = { 0x01, 0x02 };
        Event event2( 2, ConstSimpleData( buffer, sizeof(buffer) ) );

        Event event3( 3 );
        Event event4( 4 );

        SyncPointResolver syncPointResolver( *selector, *socket );

        Vector<Event> requested;
        requested.append( event1 );
        requested.append( event2 );

        Vector<Event> watched;
        watched.append( event3 );

        Vector<Event> blocked;
        blocked.append( event4 );

        char socketData1[] = { 0x05, 0x00, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0x01, 0x01 };
        ConstSimpleData constData1( socketData1, sizeof(socketData1) );

        char socketData2[] = { 0x03, 0x00, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0x01, 0x02 };
        ConstSimpleData constData2( socketData2, sizeof(socketData2) );

        String label = "hi";

        TS_ASSERT( !syncPointResolver.done() );

        TS_ASSERT_THROWS_NOTHING( syncPointResolver.go( requested, watched, blocked, IBThread::NO_TIMEOUT, noPropositions, label ) );

        TS_ASSERT( !syncPointResolver.done() );

        TS_ASSERT_THROWS_NOTHING( syncPointResolver.handleData( constData1 ) );

        TS_ASSERT( !syncPointResolver.done() );

        TS_ASSERT_THROWS_NOTHING( syncPointResolver.handleData( constData2 ) );

        TS_ASSERT( syncPointResolver.done() );
        TS_ASSERT_EQUALS( socket->lastSentData.size(), 2U );

        TS_ASSERT_EQUALS( syncPointResolver.lastEvent().type(), 0x00000003U );
        TS_ASSERT( !syncPointResolver.lastEventIsTimeout() );
        TS_ASSERT_EQUALS( syncPointResolver.lastEvent().payloadSize(), 2U );
    }

    void test_go__no_timeout__blocked_as_predicate()
    {
        Event event1( 1 );

        char buffer[] = { 0x01, 0x02 };
        Event event2( 2, ConstSimpleData( buffer, sizeof(buffer) ) );

        Event event3( 3 );
        Event event4( 4 );

        SyncPointResolver syncPointResolver( *selector, *socket );

        Vector<Event> requested;
        requested.append( event1 );
        requested.append( event2 );

        Vector<Event> watched;
        watched.append( event3 );

        BlockPredicateForTests blockPredicate;
        BlockPredicateForTests *predicatePointer = &blockPredicate;

        char pointerByte1 = ( ((int)predicatePointer) & 0x000000FF );
        char pointerByte2 = ( ( ((int)predicatePointer) >> 8 ) & 0x000000FF );
        char pointerByte3 = ( ( ((int)predicatePointer) >> 16 ) & 0x000000FF );
        char pointerByte4 = ( ( ((int)predicatePointer) >> 24 ) & 0x000000FF );

        char socketData1[] = { 0x01, 0x02, 0x03, 0x04, 0x00, 0x00, 0x00, 0x00 };
        ConstSimpleData constData1( socketData1, sizeof(socketData1) );

        char socketData2[] = { 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
        ConstSimpleData constData2( socketData2, sizeof(socketData2) );

        String label = "hi";

        TS_ASSERT( !syncPointResolver.done() );

        TS_ASSERT_THROWS_NOTHING( syncPointResolver.go( requested, watched, predicatePointer, IBThread::NO_TIMEOUT, noPropositions, label ) );

        TS_ASSERT( !syncPointResolver.done() );

        char expectedData[] = {
            // Label
            0x02, 0x00, 0x00, 0x00,
            0x68, 0x69,

            // Requested
            0x02, 0x00, 0x00, 0x00, // Number of events
            0x08, 0x00, 0x00, 0x00, // Length of first
            0x01, 0x00, 0x00, 0x00, // First
            0x00, 0x00, 0x00, 0x00,
            0x0a, 0x00, 0x00, 0x00, // Length of Second
            0x02, 0x00, 0x00, 0x00, // Second
            0x02, 0x00, 0x00, 0x00,
            0x01, 0x02,

            // Blocked
            0x01, 0x00, 0x00, 0x00, // Type = Predicate
            pointerByte1, pointerByte2, pointerByte3, pointerByte4,

            // Atomic Propositions
            0x00, 0x00, 0x00, 0x00 // Number of propositions
        };

        TS_ASSERT_EQUALS( socket->lastSentData.size(), 1U );

        TS_ASSERT_EQUALS( socket->lastSentData[0].size(), sizeof(expectedData) );
        TS_ASSERT_SAME_DATA( socket->lastSentData[0].data(), expectedData, sizeof(expectedData) );

        TS_ASSERT( !syncPointResolver.done() );

        TS_ASSERT_THROWS_NOTHING( syncPointResolver.handleData( constData1 ) );

        TS_ASSERT_EQUALS( socket->lastSentData.size(), 2U );

        TS_ASSERT_EQUALS( socket->lastSentData[1].size(), sizeof(expectedData) );
        TS_ASSERT_SAME_DATA( socket->lastSentData[1].data(), expectedData, sizeof(expectedData) );

        TS_ASSERT( !syncPointResolver.done() );

        TS_ASSERT_THROWS_NOTHING( syncPointResolver.handleData( constData2 ) );

        TS_ASSERT( syncPointResolver.done() );
        TS_ASSERT_EQUALS( socket->lastSentData.size(), 2U );

        TS_ASSERT_EQUALS( syncPointResolver.lastEvent().type(), 0x00000001U );
        TS_ASSERT( !syncPointResolver.lastEventIsTimeout() );
        TS_ASSERT_EQUALS( syncPointResolver.lastEvent().payloadSize(), 0U );
    }

    // void test_todo()
    // {
    //     TS_WARN( "BUG: if a thread timeouts and doesnt immediately re-syncs, the weaver can choose an event based on previous sync data." );
    //     // One solution: move the timeout mechanism into the weaver, and prevent race conditions
    //     // Another solution: allow threads to "withdraw" their sync data when they time out
    //     //                   For super-duper protection, might need a lock in the weaver (in case another thread simultanoulsy syncs...).
    // }
};

//
// Local Variables:
// compile-command: "make -C ../.. "
// tags-file-name: "../../TAGS"
// c-basic-offset: 4
// End:
//
