#include <cxxtest/TestSuite.h>

#include "BlockPredicateForTests.h"
#include "MockErrno.h"
#include "MockSelector.h"
#include "MockSocket.h"
#include "MockSocketFactory.h"
#include "MockWeaver.h"
#include "RealMalloc.h"
#include "ThreadListener.h"

class MockForThreadListener :
    public RealMalloc,
    public MockErrno,
    public MockSocketFactory
{
public:
};

class ThreadListenerTestSuite : public CxxTest::TestSuite
{
public:
	MockForThreadListener *mock;
    MockSelector *selector;
    MockWeaver *weaver;
    MockSocket *socket;

	void setUp()
	{
		TS_ASSERT( mock = new MockForThreadListener );
        TS_ASSERT( selector = new MockSelector );
        TS_ASSERT( weaver = new MockWeaver );

        socket = &( mock->mockSocket );
        socket->wasCreated = true;
    }

	void tearDown()
	{
        TS_ASSERT_THROWS_NOTHING( delete weaver );
        TS_ASSERT_THROWS_NOTHING( delete selector );
        TS_ASSERT_THROWS_NOTHING( delete mock );
	}

    void test_constructor_destructor()
    {
        ThreadListener *threadListener;

        TS_ASSERT( threadListener = new ThreadListener( *weaver, *selector, *socket ) );

        TS_ASSERT_EQUALS( selector->lastRegistetedDescriptor, socket->descriptor() );
        TS_ASSERT_EQUALS( selector->lastRegisteredSelectable, threadListener );

        TS_ASSERT_THROWS_NOTHING( delete threadListener );

        TS_ASSERT_EQUALS( selector->lastUnregisteredSelectable, threadListener );
        TS_ASSERT( socket->wasDiscarded );
    }

    void test_can_read__list()
    {
        ThreadListener threadListener( *weaver, *selector, *socket );

        char serialized[] = {
            // Label
            0x03, 0x00, 0x00, 0x00,
            0x68, 0x69, 0x00,

            // Requested
            0x01, 0x00, 0x00, 0x00, // Number of events
            0x0a, 0x00, 0x00, 0x00, // Length of first
            0x0a, 0x0b, 0x0c, 0x0e, // First
            0x02, 0x00, 0x00, 0x00,
            0xa1, 0xb2,

            // Blocked
            0x00, 0x00, 0x00, 0x00, // Type = List
            0x01, 0x00, 0x00, 0x00, // Number of events
            0x08, 0x00, 0x00, 0x00, // Length of first
            0x33, 0x11, 0x94, 0xa5, // First
            0x00, 0x00, 0x00, 0x00,

            // Atomic Propositions
            0x00, 0x00, 0x00, 0x00 // Number of propositions
        };

        socket->nextReceivedData.append( ConstSimpleData( serialized, sizeof(serialized) ) );

        TS_ASSERT_THROWS_NOTHING( threadListener.canRead() );

        TS_ASSERT_EQUALS( weaver->lastRequested.size(), 1U );
        TS_ASSERT_EQUALS( weaver->lastRequested[0].type(), 0x0e0c0b0aU );

        char payload[] = { 0xa1, 0xb2 };
        TS_ASSERT_EQUALS( weaver->lastRequested[0].payloadSize(), sizeof(payload) );
        TS_ASSERT_SAME_DATA( weaver->lastRequested[0].payload().data(), payload, sizeof(payload) );

        TS_ASSERT_EQUALS( weaver->lastBlocked.size(), 1U );
        TS_ASSERT_EQUALS( weaver->lastBlocked[0].type(), 0xa5941133U );
        TS_ASSERT_EQUALS( weaver->lastBlocked[0].payloadSize(), 0U );

        TS_ASSERT_EQUALS( weaver->lastThread, &threadListener );
    }

    void test_can_read__predicate()
    {
        ThreadListener threadListener( *weaver, *selector, *socket );

        BlockPredicateForTests blockPredicate;
        BlockPredicateForTests *predicatePointer = &blockPredicate;

        char pointerByte1 = ( ((int)predicatePointer) & 0x000000FF );
        char pointerByte2 = ( ( ((int)predicatePointer) >> 8 ) & 0x000000FF );
        char pointerByte3 = ( ( ((int)predicatePointer) >> 16 ) & 0x000000FF );
        char pointerByte4 = ( ( ((int)predicatePointer) >> 24 ) & 0x000000FF );

        char serialized[] = {
            // Label
            0x03, 0x00, 0x00, 0x00,
            0x68, 0x69, 0x00,

            // Requested
            0x01, 0x00, 0x00, 0x00, // Number of events
            0x0a, 0x00, 0x00, 0x00, // Length of first
            0x0a, 0x0b, 0x0c, 0x0e, // First
            0x02, 0x00, 0x00, 0x00,
            0xa1, 0xb2,

            // Blocked
            0x01, 0x00, 0x00, 0x00, // Type = Predicate
            pointerByte1, pointerByte2, pointerByte3, pointerByte4,

            // Atomic Propositions
            0x00, 0x00, 0x00, 0x00 // Number of propositions
        };

        socket->nextReceivedData.append( ConstSimpleData( serialized, sizeof(serialized) ) );

        TS_ASSERT_THROWS_NOTHING( threadListener.canRead() );

        TS_ASSERT_EQUALS( weaver->lastRequested.size(), 1U );
        TS_ASSERT_EQUALS( weaver->lastRequested[0].type(), 0x0e0c0b0aU );

        char payload[] = { 0xa1, 0xb2 };
        TS_ASSERT_EQUALS( weaver->lastRequested[0].payloadSize(), sizeof(payload) );
        TS_ASSERT_SAME_DATA( weaver->lastRequested[0].payload().data(), payload, sizeof(payload) );

        TS_ASSERT_EQUALS( weaver->lastBlockPredicate, predicatePointer );

        TS_ASSERT_EQUALS( weaver->lastThread, &threadListener );
    }

    void test_annouce_chosen_event()
    {
        ThreadListener threadListener( *weaver, *selector, *socket );

        char serialized[] = {
            // Label
            0x03, 0x00, 0x00, 0x00,
            0x68, 0x69, 0x00,

            // Requested
            0x01, 0x00, 0x00, 0x00, // Number of events
            0x0c, 0x00, 0x00, 0x00, // Length of first
            0x0a, 0x0b, 0x0c, 0x0e, // First
            0x04, 0x00, 0x00, 0x00,
            0xab, 0xcd, 0xef, 0x12,

            // Blocked
            0x00, 0x00, 0x00, 0x00, // Type = List
            0x01, 0x00, 0x00, 0x00, // Number of events
            0x08, 0x00, 0x00, 0x00, // Length of first
            0x33, 0x11, 0x94, 0xa5, // First
            0x00, 0x00, 0x00, 0x00,

            // Atomic Propositions
            0x00, 0x00, 0x00, 0x00 // Number of propositions
        };

        socket->nextReceivedData.append( ConstSimpleData( serialized, sizeof(serialized) ) );
        TS_ASSERT_THROWS_NOTHING( threadListener.canRead() );

        char buffer[] = { 0x12, 0x34 };

        Event announcedEvent( 0x04030201, ConstSimpleData( buffer, sizeof(buffer) ) );
        char serializedEvent[] = { 0x01, 0x02, 0x03, 0x04,
                                   0x02, 0x00, 0x00, 0x00,
                                   0x12, 0x34 };

        TS_ASSERT_THROWS_NOTHING( threadListener.announceChosenEvent( announcedEvent ) );

        TS_ASSERT_EQUALS( socket->lastSentData.size(), 1U );
        TS_ASSERT_EQUALS( socket->lastSentData[0].size(), sizeof(serializedEvent) );
        TS_ASSERT_SAME_DATA( socket->lastSentData[0].data(), serializedEvent, sizeof(serializedEvent) );
    }

    void test_can_read__socket_closed()
    {
        ThreadListener threadListener( *weaver, *selector, *socket );

        socket->socketShouldClose = true;

        TS_ASSERT_THROWS_NOTHING( threadListener.canRead() );

        TS_ASSERT( weaver->removeThreadWasCalled );
        TS_ASSERT( !weaver->bSyncWasCalled );
        TS_ASSERT_EQUALS( weaver->lastRemovedThread, &threadListener );
    }

    void test_can_read__socket_error()
    {
        ThreadListener threadListener( *weaver, *selector, *socket );

        socket->socketShouldThrow = true;

        TS_ASSERT_THROWS_EQUALS( threadListener.canRead(),
                                 const Error &e,
                                 e.code(),
                                 Error::RECV_FAILED );

        TS_ASSERT( !weaver->removeThreadWasCalled );
    }


    void test_annouce_chosen_event__before_can_read()
    {
        ThreadListener threadListener( *weaver, *selector, *socket );

        char serialized[] = {
            // Label
            0x03, 0x00, 0x00, 0x00,
            0x68, 0x69, 0x00,

            // Requested
            0x01, 0x00, 0x00, 0x00, // Number of events
            0x0c, 0x00, 0x00, 0x00, // Length of first
            0x0a, 0x0b, 0x0c, 0x0e, // First
            0x04, 0x00, 0x00, 0x00,
            0xab, 0xcd, 0xef, 0x12,

            // Blocked
            0x00, 0x00, 0x00, 0x00, // Type = List
            0x01, 0x00, 0x00, 0x00, // Number of events
            0x08, 0x00, 0x00, 0x00, // Length of first
            0x33, 0x11, 0x94, 0xa5, // First
            0x00, 0x00, 0x00, 0x00,

            // Atomic Propositions
            0x00, 0x00, 0x00, 0x00 // Number of propositions
        };

        char buffer[] = { 0x12, 0x34 };

        Event announcedEvent1( 0x04030201, ConstSimpleData( buffer, sizeof(buffer) ) );
        char serializedEvent1[] = { 0x01, 0x02, 0x03, 0x04,
                                   0x02, 0x00, 0x00, 0x00,
                                   0x12, 0x34 };

        Event announcedEvent2( 2 );
        char serializedEvent2[] = { 0x02, 0x00, 0x00, 0x00,
                                    0x00, 0x00, 0x00, 0x00 };

        TS_ASSERT_THROWS_NOTHING( threadListener.announceChosenEvent( announcedEvent1 ) );
        TS_ASSERT_EQUALS( socket->lastSentData.size(), 0U );

        TS_ASSERT_THROWS_NOTHING( threadListener.announceChosenEvent( announcedEvent2 ) );
        TS_ASSERT_EQUALS( socket->lastSentData.size(), 0U );

        socket->nextReceivedData.append( ConstSimpleData( serialized, sizeof(serialized) ) );
        TS_ASSERT_THROWS_NOTHING( threadListener.canRead() );

        TS_ASSERT_EQUALS( socket->lastSentData.size(), 1U );
        TS_ASSERT_EQUALS( socket->lastSentData[0].size(), sizeof(serializedEvent1) );
        TS_ASSERT_SAME_DATA( socket->lastSentData[0].data(), serializedEvent1, sizeof(serializedEvent1) );
        TS_ASSERT_EQUALS( weaver->lastRequested.size(), 0U );

        socket->nextReceivedData.append( ConstSimpleData( serialized, sizeof(serialized) ) );
        TS_ASSERT_THROWS_NOTHING( threadListener.canRead() );

        TS_ASSERT_EQUALS( socket->lastSentData.size(), 2U );
        TS_ASSERT_EQUALS( socket->lastSentData[1].size(), sizeof(serializedEvent2) );
        TS_ASSERT_SAME_DATA( socket->lastSentData[1].data(), serializedEvent2, sizeof(serializedEvent2) );
        TS_ASSERT_EQUALS( weaver->lastRequested.size(), 0U );

        socket->nextReceivedData.append( ConstSimpleData( serialized, sizeof(serialized) ) );
        TS_ASSERT_THROWS_NOTHING( threadListener.canRead() );
        TS_ASSERT_EQUALS( socket->lastSentData.size(), 2U );
        TS_ASSERT_EQUALS( weaver->lastRequested.size(), 1U );
    }
};

//
// Local Variables:
// compile-command: "make -C ../.. "
// tags-file-name: "../../TAGS"
// c-basic-offset: 4
// End:
//
