#ifndef __SegmentSorter_h__
#define __SegmentSorter_h__

#include "BThread.h"
#include "IEvent.h"
#include "SortedList.h"
#include "TcpSegment.h"

class SegmentSorter : public BThread
{
public:
    struct StoredPacket
    {
        unsigned _sequenceNumber;
        unsigned _size;
        HeapData _payload;

        StoredPacket( unsigned sequenceNumber, unsigned size, const ConstSimpleData &payload ) : _sequenceNumber( sequenceNumber ), _size( size ), _payload( payload )
        {
        }

        bool operator<( const StoredPacket &other )
        {
            return _sequenceNumber < other._sequenceNumber;
        }

        StoredPacket &operator=( const StoredPacket &other )
        {
            _sequenceNumber = other._sequenceNumber;
            _size = other._size;
            _payload = ConstSimpleData( other._payload.data(), other._payload.size() );

            return *this;
        }
    };

    SegmentSorter() : BThread( BThread::REGULAR_THREAD )
    {
    }

    void entryPoint()
    {
        try
        {
            while ( true )
            {
                waitForDataAndSyns();

                if ( synRequest() )
                {
                    log( "Handling syn request" );
                    handleSynRequest();
                }
                else
                {
                    extractRequest();
                    addPacketToList();
                    dispatchPacketsIfPossible();
                }
            }
        }
        catch ( Error &e )
        {
            e.setUserMessage( "Thread: SegmentSorter" );
            throw e;
        }
    }

private:
    typedef SortedList<StoredPacket> SortedPacketList;

    Map<ConnectionIdentifier, unsigned> _connectionToNextByte;
    Map<ConnectionIdentifier, SortedPacketList> _connectionToPackets;

    TcpEvent _tcpEvent;
    TcpSegment _tcpSegment;
    ConnectionIdentifier _connectionIdentifier;

    bool synRequest() const
    {
        return lastEvent().type() == IEvent::TCP_SYN_REQUEST;
    }

    void handleSynRequest()
    {
        _tcpEvent.constructFromEvent( lastEvent() );
        _tcpSegment.unserialize( _tcpEvent.getPacketPayload() );
        _connectionIdentifier = ConnectionIdentifier( _tcpEvent.getIp(), _tcpEvent.getPort() );

        _connectionToNextByte[_connectionIdentifier] = _tcpSegment.header().getSequenceNumber() + 1;

        log( Stringf( "next sequence number for connection (%s, %u): %x", _tcpEvent.getIp().ascii(), _tcpEvent.getPort(), _tcpSegment.header().getSequenceNumber() + 1 ) );
    }

    void waitForDataAndSyns()
    {
        clearEventVectors();
        _watched.append( IEvent::TCP_SYN_REQUEST );
        _watched.append( IEvent::DATA_FROM_TCP_TO_HTTP );

        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    void extractRequest()
    {
        _tcpEvent.constructFromEvent( lastEvent() );
        _connectionIdentifier = ConnectionIdentifier( _tcpEvent.getIp(), _tcpEvent.getPort() );
        _tcpSegment.unserialize( _tcpEvent.getPacketPayload() );
    }

    void addPacketToList()
    {
        if ( !_connectionToPackets.exists( _connectionIdentifier ) )
        {
            SortedPacketList sortedPacketList;
            _connectionToPackets[_connectionIdentifier] = sortedPacketList;
        }

        _connectionToPackets[_connectionIdentifier].insert( StoredPacket( _tcpSegment.header().getSequenceNumber(), _tcpSegment._payload.size(), _tcpSegment._payload ) );
    }

    void dispatchPacketsIfPossible()
    {
        SortedPacketList::iterator it = _connectionToPackets[_connectionIdentifier].begin();
        unsigned &currentSequenceNumber( _connectionToNextByte[_connectionIdentifier] );

        log( Stringf( "Current sequence number is: %x", currentSequenceNumber ) );

        bool flag = false;

        while ( ( it != _connectionToPackets[_connectionIdentifier].end() ) && ( !flag ) )
        {
            if ( it->_sequenceNumber < currentSequenceNumber )
            {
                it = _connectionToPackets[_connectionIdentifier].erase( it );
            }
            else if ( it->_sequenceNumber == currentSequenceNumber )
            {
                log( Stringf( "Dispatching packet with sequence number: %x", it->_sequenceNumber ) );

                sendPacket( *it );
                currentSequenceNumber = it->_sequenceNumber + it->_size;
                it = _connectionToPackets[_connectionIdentifier].erase( it );

                log( Stringf( "Current sequence number is: %x", currentSequenceNumber ) );
            }
            else
            {
                flag = true;
            }
        }

        if ( !flag )
        {
            log( "All packets dispatched!\n" );
        }
        else
        {
            log( Stringf( "%u packets remain...\n", _connectionToPackets[_connectionIdentifier].size() ) );
        }
    }

    void sendPacket( StoredPacket &packet )
    {
        clearEventVectors();
        _requested.append( TcpEvent( IEvent::ORDERED_DATA_FROM_HTTP_TO_TCP,
                                     ipUnsignedToString( _connectionIdentifier._ip ),
                                     _connectionIdentifier._port,
                                     packet._payload ) );

        _blocked.append( IEvent::DATA_FROM_TCP_TO_HTTP );

        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    void log( String message )
    {
        return;
        printf( "\t\t\tSegmentSorter: %s\n", message.ascii() );
    }

    static String ipUnsignedToString( unsigned ip )
    {
        struct in_addr ip_addr;
        memset( &ip_addr, 0, sizeof(ip_addr) );
        ip_addr.s_addr = ntohl( ip );
        return inet_ntoa( ip_addr );
    }
};

#endif // __SegmentSorter_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
