#ifndef __TcpAckSender_h__
#define __TcpAckSender_h__

#include "BThread.h"
#include "IEvent.h"
#include "ServerConfiguration.h"
#include "TcpSegment.h"

class TcpAckSender : public BThread
{
public:
    TcpAckSender() : BThread( BThread::REGULAR_THREAD )
    {
    }

    void entryPoint()
    {
        try
        {
            while ( true )
            {
                waitForAckRequest();
                storeSegmentToAck();
                getFreshSequenceNumber();
                updateAckNumber();
                sendAck();
                doneHandlingAck();
            }
        }
        catch ( Error &e )
        {
            e.setUserMessage( "Thread: TcpAckSender" );
            throw e;
        }
    }

private:
    unsigned _sequenceNumber;
    TcpEvent _ackRequest;
    TcpSegment _segmentToAck;

    void storeSegmentToAck()
    {
        _ackRequest.constructFromEvent( lastEvent() );
        _segmentToAck.unserialize( _ackRequest.getPacketPayload() );
    }

    void waitForAckRequest()
    {
        clearEventVectors();

        _watched.append( IEvent::TCP_SEND_ACK );
        _blocked = allRelevantEventsExcept( IEvent::TCP_SEND_ACK );
        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    void getFreshSequenceNumber()
    {
        requestSequenceNumber();
        readSequenceNumber();
    }

    void requestSequenceNumber()
    {
        enum {
            DONT_CARE = 0,
            ACKS_DONT_HAVE_PAYLOADS = 0,
        };

        clearEventVectors();

        SequenceNumberPayload sequenceNumber( DONT_CARE, ACKS_DONT_HAVE_PAYLOADS );

        _requested.append( TcpEvent( IEvent::ACK_SENDER_REQUEST_SEQUENCE_NUMBER,
                                     _ackRequest.getIp(),
                                     _ackRequest.getPort(),
                                     ConstSimpleData( &sequenceNumber, sizeof(sequenceNumber) ) ) );
        _blocked = allRelevantEventsExcept( IEvent::ACK_SENDER_REQUEST_SEQUENCE_NUMBER );
        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    void readSequenceNumber()
    {
        clearEventVectors();

        _watched.append( IEvent::SEQUENCE_NUMBER_FOR_ACK_SENDER );
        _blocked = allRelevantEventsExcept( IEvent::SEQUENCE_NUMBER_FOR_ACK_SENDER );
        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );

        TcpEvent sequenceNumber;
        sequenceNumber.constructFromEvent( lastEvent() );

        _sequenceNumber = SequenceNumberPayload( sequenceNumber.getPacketPayload() ).getSequenceNumber();
    }

    void updateAckNumber()
    {
        AckNumberPayload payload( _segmentToAck.header().getSequenceNumber() + _segmentToAck._payload.size() );

        clearEventVectors();
        _requested.append( TcpEvent( IEvent::SET_ACK_NUMBER, _ackRequest.getIp(), _ackRequest.getPort(), ConstSimpleData( &payload, sizeof(payload) ) ) );
        _blocked = allRelevantEventsExcept( IEvent::SET_ACK_NUMBER );
        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    void sendAck()
    {
        TcpSegment ack;
        prepareAck( ack );
        sendAck( ack );
    }

    void prepareAck( TcpSegment &ack )
    {
        ack.header().setSourcePort( _segmentToAck.header().getDestinationPort() );
        ack.header().setDestinationPort( _segmentToAck.header().getSourcePort() );

        ack.header().setAckNumber( _segmentToAck.header().getSequenceNumber() + _segmentToAck._payload.size() );
        ack.header().setSequenceNumber( _sequenceNumber );

        ack.header().setAck();

        ack.header().setWindowSize( ServerConfiguration::TCP_WINDOW_SIZE );
    }

    void sendAck( TcpSegment &synAck )
    {
        HeapData serialized;
        synAck.serialize( serialized, ServerConfiguration::SERVER_IP, _ackRequest.getIp() );

        clearEventVectors();
        IpEvent ackEvent( IEvent::TCP_OUTGOING_SEGMENT, _ackRequest.getIp(),  serialized );
        _requested.append( ackEvent );

        _blocked = allRelevantEventsExcept( IEvent::TCP_OUTGOING_SEGMENT );
        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    void doneHandlingAck()
    {
        clearEventVectors();
        _requested.append( IEvent::TCP_DONE_HADNLING_ACK );
        _blocked = allRelevantEventsExcept( IEvent::TCP_DONE_HADNLING_ACK );
        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    Vector<Event> allRelevantEventsExcept( Event event )
    {
        Vector<Event> result;

        if ( event != IEvent::TCP_SEND_ACK ) result.append( IEvent::TCP_SEND_ACK );
        if ( event != IEvent::ACK_SENDER_REQUEST_SEQUENCE_NUMBER ) result.append( IEvent::ACK_SENDER_REQUEST_SEQUENCE_NUMBER );
        if ( event != IEvent::SEQUENCE_NUMBER_FOR_ACK_SENDER ) result.append( IEvent::SEQUENCE_NUMBER_FOR_ACK_SENDER );
        if ( event != IEvent::TCP_DONE_HADNLING_ACK ) result.append( IEvent::TCP_DONE_HADNLING_ACK );

        return result;
    }
};

#endif // __TcpAckSender_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
