#ifndef __TcpDispatcher_h__
#define __TcpDispatcher_h__

#include "BThread.h"
#include "IEvent.h"
#include "TcpSegment.h"

class TcpDispatcher : public BThread
{
public:
    TcpDispatcher() : BThread( BThread::REGULAR_THREAD )
    {
    }

    void entryPoint()
    {
        try
        {
            while ( true )
            {
                waitForRequests();
                classifyRequest();
            }
        }
        catch ( Error &e )
        {
            e.setUserMessage( "Thread: TcpDispatcher" );
            throw e;
        }
    }

private:
    IpEvent _ipEvent;
    TcpSegment _tcpSegment;

    void extractIpEvent()
    {
        _ipEvent.constructFromEvent( lastEvent() );
    }

    void extractTcpSegment()
    {
        memset( &_tcpSegment, 0, sizeof(TcpSegment) );
        _tcpSegment.unserialize( _ipEvent.getPacketPayload() );
    }

    void waitForRequests()
    {
        clearEventVectors();
        _watched.append( IEvent::TCP_SEGMENT_RECEIVED );
        _blocked = allRelevantEventsExcept( IEvent::TCP_SEGMENT_RECEIVED );
        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    bool packetForUs()
    {
        return _tcpSegment.header().getDestinationPort() == ServerConfiguration::LISTENING_PORT;
    }

    void classifyRequest()
    {
        log( "Packet Received" );

        extractIpEvent();
        extractTcpSegment();

        if ( !packetForUs() )
        {
            return;
        }

        bool handled = false;

        if ( _tcpSegment.header().getSyn() )
        {
            log( "Syn" );
            synRequest();
            handled = true;
        }
        if ( _tcpSegment.header().getAck() )
        {
            log( "Ack" );
            handleAck();
            handled = true;
        }
        if ( _tcpSegment._payload.size() > 0 )
        {
            log( "Data" );
            handlePushedData();
            handled = true;
        }
        if ( _tcpSegment.header().getFin() )
        {
            log( "Fin" );
            finRequest();
            handled = true;
        }

        if ( !handled )
        {
            log( "Unknown packet!\n" );
        }
    }

    void synRequest()
    {
        clearEventVectors();
        _requested.append( TcpEvent( IEvent::TCP_SYN_REQUEST,
                                     _ipEvent.getIp(),
                                     _tcpSegment.header().getSourcePort(),
                                     _ipEvent.getPacketPayload() ) );
        _requested.append( IEvent::IDLE_EVENT );

        _blocked = allRelevantEventsExcept( IEvent::TCP_SYN_REQUEST );

        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    void finRequest()
    {
        clearEventVectors();
        _requested.append( TcpEvent( IEvent::TCP_FIN_REQUEST,
                                     _ipEvent.getIp(),
                                     _tcpSegment.header().getSourcePort(),
                                     _ipEvent.getPacketPayload() ) );
        _requested.append( IEvent::IDLE_EVENT );

        _blocked = allRelevantEventsExcept( IEvent::TCP_FIN_REQUEST );

        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    void handlePushedData()
    {
        clearEventVectors();
        _requested.append( TcpEvent( IEvent::TCP_PUSH_REQUEST,
                                     _ipEvent.getIp(),
                                     _tcpSegment.header().getSourcePort(),
                                     _ipEvent.getPacketPayload() ) );
        _requested.append( IEvent::IDLE_EVENT );

        _blocked = allRelevantEventsExcept( IEvent::TCP_PUSH_REQUEST );
        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    void handleAck()
    {
        clearEventVectors();
        _requested.append( TcpEvent( IEvent::TCP_ACK_REQUEST,
                                     _ipEvent.getIp(),
                                     _tcpSegment.header().getSourcePort(),
                                     _ipEvent.getPacketPayload() ) );
        _requested.append( IEvent::IDLE_EVENT );

        _blocked = allRelevantEventsExcept( IEvent::TCP_ACK_REQUEST );
        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    Vector<Event> allRelevantEventsExcept( Event event )
    {
        Vector<Event> result;

        if ( event != IEvent::TCP_SYN_REQUEST ) result.append( IEvent::TCP_SYN_REQUEST );
        if ( event != IEvent::TCP_FIN_REQUEST ) result.append( IEvent::TCP_FIN_REQUEST );
        if ( event != IEvent::TCP_PUSH_REQUEST ) result.append( IEvent::TCP_PUSH_REQUEST );
        if ( event != IEvent::TCP_SEGMENT_RECEIVED ) result.append( IEvent::TCP_SEGMENT_RECEIVED );
        if ( event != IEvent::TCP_ACK_REQUEST ) result.append( IEvent::TCP_ACK_REQUEST );

        return result;
    }

    void log( String message )
    {
        return;
        printf( "TcpDispatcher:\t%s\n", message.ascii() );
    }
};

#endif // __TcpDispatcher_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
