#ifndef __TcpEnsureActiveConnection_h__
#define __TcpEnsureActiveConnection_h__

#include "BThread.h"
#include "EventPredicate.h"
#include "IEvent.h"

class TcpEnsureActiveConnection : public BThread
{
public:
    TcpEnsureActiveConnection() : BThread( BThread::REGULAR_THREAD )
    {
    }

    class BlockUnlessDataTransfer : public EventPredicate
    {
    public:
        bool evaluate( const Event &event )
        {
            if ( ( event.type() != IEvent::TCP_FIN_REQUEST ) &&
                 ( event.type() != IEvent::TCP_PUSH_REQUEST ) &&
                 ( event.type() != IEvent::TCP_ACK_REQUEST ) &&
                 ( event.type() != IEvent::TCP_SYN_REQUEST ) )
            {
                return false;
            }

            TcpEvent tcpEvent;
            tcpEvent.constructFromEvent( event );
            ConnectionIdentifier connectionIdentifier( tcpEvent.getIp(), tcpEvent.getPort() );

            if ( event.type() == IEvent::TCP_SYN_REQUEST )
            {
                return _activeConnections.exists( connectionIdentifier );
            }

            return !_activeConnections.exists( connectionIdentifier );
        }

        void addConnection( const ConnectionIdentifier &connectionIdentifier )
        {
            // printf( "Adding active connection: port %u\n", connectionIdentifier._port );
            _activeConnections.insert( _activeConnections.begin(), connectionIdentifier );
        }

        void removeConnection( const ConnectionIdentifier &connectionIdentifier )
        {
            try
            {
                // printf( "Removing connection: port %u\n", connectionIdentifier._port );
                _activeConnections.eraseByValue( connectionIdentifier );
            }
            catch ( const Error &e )
            {
                if ( e.code() != Error::VALUE_DOESNT_EXIST_IN_LIST )
                {
                    throw;
                }
            }
        }

    private:
        List<ConnectionIdentifier> _activeConnections;
    };

    void entryPoint()
    {
        try
        {
            while ( true )
            {
                waitForSynsAndFins();
                handleLastEvent();
            }
        }
        catch ( Error &e )
        {
            e.setUserMessage( "Thread: TcpEnsureActiveConnection" );
            throw e;
        }
    }

private:
    BlockUnlessDataTransfer _blockUnlessDataTransfer;

    void waitForSynsAndFins()
    {
        _watched.append( IEvent::TCP_SYN_REQUEST );
        _watched.append( IEvent::TCP_FIN_REQUEST );

        bSync( _requested, _watched, &_blockUnlessDataTransfer, IBThread::NO_TIMEOUT );
    }

    void handleLastEvent()
    {
        TcpEvent tcpEvent;
        tcpEvent.constructFromEvent( lastEvent() );
        ConnectionIdentifier connectionIdentifier( tcpEvent.getIp(), tcpEvent.getPort() );

        if ( lastEvent().type() == IEvent::TCP_SYN_REQUEST )
        {
            _blockUnlessDataTransfer.addConnection( connectionIdentifier );
        }
        else
        {
            _blockUnlessDataTransfer.removeConnection( connectionIdentifier );
        }
    }
};

#endif // __TcpEnsureActiveConnection_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
