#ifndef __TcpKeepAliveSender_h__
#define __TcpKeepAliveSender_h__

#include "ConnectionIdentifier.h"
#include "ITcpKeepAliveMaster.h"
#include "IpEvent.h"
#include "SequenceNumberPayload.h"
#include "ServerConfiguration.h"
#include "SpawnableBThread.h"
#include "TcpSegment.h"

class TcpKeepAliveSender : public SpawnableBThread
{
public:
    TcpKeepAliveSender( ConnectionIdentifier connectionIdentifier, ITcpKeepAliveMaster &master ) :
        SpawnableBThread( BThread::REGULAR_THREAD ), _connectionIdentifier( connectionIdentifier ), _master( master )
    {
    }

    void entryPoint()
    {
        storeTime();

        while ( !_master.shouldTerminate( _connectionIdentifier ) )
        {
            monitorConnectionActivity();

            if ( shouldSendKeepAlive() )
            {
                sendKeepAlive();
            }
        }

        done();
    }

private:
    ConnectionIdentifier _connectionIdentifier;
    ITcpKeepAliveMaster &_master;
    time_t _lastActivityTime;
    unsigned _sequenceNumber;
    unsigned _ackNumber;

    enum {
        SLEEP_INCREMENT_SECONDS = 1,
    };

    void monitorConnectionActivity()
    {
        clearEventVectors();
        _watched.append( IEvent::TCP_PUSH_REQUEST );

        if ( bSync( _requested, _watched, _blocked, SLEEP_INCREMENT_SECONDS ) == IBThread::EVENT_WAS_TRIGGERED )
        {
            TcpEvent pushRequest;
            pushRequest.constructFromEvent( lastEvent() );

            ConnectionIdentifier requestIdentifier( pushRequest.getIp(), pushRequest.getPort() );

            if ( _connectionIdentifier == requestIdentifier )
            {
                storeTime();
            }
        }
    }

    void storeTime()
    {
        time_t *DONT_CARE = NULL;
        _lastActivityTime = time( DONT_CARE );
    }

    bool shouldSendKeepAlive() const
    {
        time_t *DONT_CARE = NULL;
        return ( time( DONT_CARE ) - _lastActivityTime > ServerConfiguration::SECONDS_FOR_KEEP_ALIVE );
    }

    void sendKeepAlive()
    {
        clearEventVectors();

        getFreshSequenceNumber();
        getAckNumber();
        sendKeepAliveMessage();

        storeTime();
    }

    void getFreshSequenceNumber()
    {
        requestSequenceNumber();
        readSequenceNumber();
    }

    void requestSequenceNumber()
    {
        enum {
            DONT_CARE = 0,
            KEEP_ALIVE_DOESNT_ADVANCE_COUNTER = 0,
        };

        clearEventVectors();

        SequenceNumberPayload sequenceNumber( DONT_CARE, KEEP_ALIVE_DOESNT_ADVANCE_COUNTER );

        _requested.append( TcpEvent( IEvent::KEEPALIVE_SENDER_REQUEST_SEQUENCE_NUMBER,
                                     ipUnsignedToString( _connectionIdentifier._ip ),
                                     _connectionIdentifier._port,
                                     ConstSimpleData( &sequenceNumber, sizeof(sequenceNumber) ) ) );
        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    void readSequenceNumber()
    {
        clearEventVectors();

        _watched.append( IEvent::SEQUENCE_NUMBER_FOR_KEEPALIVE_SENDER );
        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );

        TcpEvent sequenceNumber;
        sequenceNumber.constructFromEvent( lastEvent() );

        _sequenceNumber = SequenceNumberPayload( sequenceNumber.getPacketPayload() ).getSequenceNumber();
    }

    void getAckNumber()
    {
        sendAckRequest();
        getAckReply();
    }

    void sendAckRequest()
    {
        clearEventVectors();

        ConstSimpleData empty( 0, 0 );

        _requested.append( TcpEvent( IEvent::ACK_NUMBER_REQUEST, ipUnsignedToString( _connectionIdentifier._ip ), _connectionIdentifier._port, empty ) );
        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    void getAckReply()
    {
        TcpEvent ackReply;

        clearEventVectors();

        do
        {
            _watched.append( IEvent::ACK_NUMBER_REPLY );
            bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
            ackReply.constructFromEvent( lastEvent() );
        }
        while ( ( ackReply.getIp() != ipUnsignedToString( _connectionIdentifier._ip ) ) || ( ackReply.getPort() != _connectionIdentifier._port ) );

        _ackNumber = AckNumberPayload( ackReply.getPacketPayload() ).getAckNumber();
    }

    void sendKeepAliveMessage()
    {
        TcpSegment keepAlive;
        prepareKeepAlive( keepAlive );
        send( keepAlive );
    }

    void prepareKeepAlive( TcpSegment &keepAlive )
    {
        keepAlive.header().setSourcePort( ServerConfiguration::LISTENING_PORT );
        keepAlive.header().setDestinationPort( _connectionIdentifier._port );

        keepAlive.header().setAckNumber( _ackNumber );
        keepAlive.header().setSequenceNumber( _sequenceNumber - 1 );

        keepAlive.header().setAck();

        keepAlive.header().setWindowSize( ServerConfiguration::TCP_WINDOW_SIZE );
    }

    void send( TcpSegment &keepAlive )
    {
        HeapData serialized;
        keepAlive.serialize( serialized, ServerConfiguration::SERVER_IP, ipUnsignedToString( _connectionIdentifier._ip ) );

        clearEventVectors();
        IpEvent keepAliveEvent( IEvent::TCP_OUTGOING_SEGMENT, ipUnsignedToString( _connectionIdentifier._ip ),  serialized );
        _requested.append( keepAliveEvent );

        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    static String ipUnsignedToString( unsigned ip )
    {
        struct in_addr ip_addr;
        memset( &ip_addr, 0, sizeof(ip_addr) );
        ip_addr.s_addr = ntohl( ip );
        return inet_ntoa( ip_addr );
    }
};

#endif // __TcpKeepAliveSender_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
