#ifndef __TcpPushSender_h__
#define __TcpPushSender_h__

#include "BThread.h"
#include "IEvent.h"
#include "TargetPortPayload.h"

class TcpPushSender : public BThread
{
public:
    TcpPushSender() : BThread( BThread::REGULAR_THREAD )
    {
    }

    void entryPoint()
    {
        try
        {
            while ( true )
            {
                waitForPushRequest();
                handlePushRequest();
            }
        }
        catch ( Error &e )
        {
            e.setUserMessage( "Thread: TcpPushSender" );
            throw e;
        }
    }

private:
    unsigned _sequenceNumber;
    TcpEvent _pushRequest;

    void waitForPushRequest()
    {
        clearEventVectors();
        _watched.append( IEvent::TCP_SEND_PUSH );

        _blocked = allRelevantEventsExcept( IEvent::TCP_SEND_PUSH );

        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    void handlePushRequest()
    {
        storePushRequest();
        getFreshSequenceNumber();
        sendPacket();
    }

    void storePushRequest()
    {
        _pushRequest.constructFromEvent( lastEvent() );
    }

    void getFreshSequenceNumber()
    {
        requestSequenceNumber();
        readSequenceNumber();
    }

    void requestSequenceNumber()
    {
        enum {
            DONT_CARE = 0,
        };

        clearEventVectors();

        SequenceNumberPayload sequenceNumber( DONT_CARE, _pushRequest.getPacketPayload().size() );

        _requested.append( TcpEvent( IEvent::PUSH_SENDER_REQUEST_SEQUENCE_NUMBER,
                                     _pushRequest.getIp(),
                                     _pushRequest.getPort(),
                                     ConstSimpleData( &sequenceNumber, sizeof(sequenceNumber) ) ) );
        _blocked = allRelevantEventsExcept( IEvent::PUSH_SENDER_REQUEST_SEQUENCE_NUMBER );
        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    void readSequenceNumber()
    {
        clearEventVectors();

        _watched.append( IEvent::SEQUENCE_NUMBER_FOR_PUSH_SENDER );
        _blocked = allRelevantEventsExcept( IEvent::SEQUENCE_NUMBER_FOR_PUSH_SENDER );
        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );

        TcpEvent sequenceNumber;
        sequenceNumber.constructFromEvent( lastEvent() );

        _sequenceNumber = SequenceNumberPayload( sequenceNumber.getPacketPayload() ).getSequenceNumber();
    }

    void sendPacket()
    {
        TcpSegment push;
        preparePush( push );
        sendPush( push );
    }

    void preparePush( TcpSegment &push )
    {
        push._payload = _pushRequest.getPacketPayload();

        push.header().setSourcePort( ServerConfiguration::LISTENING_PORT );
        push.header().setDestinationPort( _pushRequest.getPort() );

        push.header().setAckNumber( 0 );
        push.header().setSequenceNumber( _sequenceNumber );

        push.header().setPush();

        push.header().setWindowSize( ServerConfiguration::TCP_WINDOW_SIZE );
    }

    void sendPush( TcpSegment &push )
    {
        HeapData serialized;
        push.serialize( serialized, ServerConfiguration::SERVER_IP, _pushRequest.getIp() );

        clearEventVectors();
        IpEvent request( IEvent::TCP_OUTGOING_SEGMENT, _pushRequest.getIp(), serialized );
        _requested.append( request );

        _blocked = allRelevantEventsExcept( IEvent::TCP_OUTGOING_SEGMENT );
        bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );

        // clearEventVectors();
        // TcpEvent storageRequest( IEvent::TCP_STORE_PACKET, _pushRequest.getIp(), _pushRequest.getPort(), serialized );
        // _requested.append( storageRequest );

        // _blocked = allRelevantEventsExcept( IEvent::TCP_STORE_PACKET );
        // bSync( _requested, _watched, _blocked, IBThread::NO_TIMEOUT );
    }

    Vector<Event> allRelevantEventsExcept( Event event )
    {
        Vector<Event> result;

        if ( event != IEvent::TCP_SEND_PUSH ) result.append( IEvent::TCP_SEND_PUSH );
        if ( event != IEvent::PUSH_SENDER_REQUEST_SEQUENCE_NUMBER ) result.append( IEvent::PUSH_SENDER_REQUEST_SEQUENCE_NUMBER );
        if ( event != IEvent::SEQUENCE_NUMBER_FOR_PUSH_SENDER ) result.append( IEvent::SEQUENCE_NUMBER_FOR_PUSH_SENDER );

        return result;
    }
};

#endif // __TcpPushSender_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
