#ifndef __Proxy_h__
#define __Proxy_h__

#include "ClientAllowAllPolicy.h"
#include "IpHeader.h"
#include "OccasionalRetransmitPolicy.h"
#include "OccasionalShufflePolicy.h"
#include "ProxyConfiguration.h"
#include "RawSocket.h"
#include "ServerAllowAllPolicy.h"
#include "ServerConfiguration.h"
#include "Stringf.h"
#include "TcpSegment.h"

class Proxy
{
public:
    Proxy() : _rawSocket( ProxyConfiguration::PROXY_IP )
    {
    }

    static void start()
    {
        Proxy().run();
    }

    void run()
    {
        log( "Proxy Starting" );

        initializeSocket();

        while ( true )
        {
            receiveData();
            handlePackets();
        }
    }

private:
    RawSocket _rawSocket;
    HeapData _data;
    struct sockaddr_in _address;
    unsigned _indexInData;

    // Server Policy
    OccasionalRetransmitPolicy _serverPacketPolicy;

    // Client Policy
    OccasionalShufflePolicy _clientPacketPolicy;

    TcpSegment _tcpSegment;
    IpHeader _ipHeader;

    Vector<HeapData> _packetsToServer;

    void receiveData()
    {
        memset( &_address, 0, sizeof(_address) );
        _rawSocket.receive( _data, _address );
    }

    void handlePackets()
    {
        _indexInData = 0;

        while ( _indexInData < _data.size() )
        {
            handleSinglePacket();
        }
    }

    void handleSinglePacket()
    {
        _ipHeader.unserialize( ConstSimpleData( _data.asChar() + _indexInData, _data.size() - _indexInData ) );
        _tcpSegment.unserialize( ConstSimpleData( _data.asChar() + _indexInData + _ipHeader.size(), _ipHeader._mandatoryPart._totalLength - _ipHeader.size() ) );
        _indexInData += _ipHeader._mandatoryPart._totalLength;

        if ( packetToServer() )
        {
            log( "Forwarding to server" );
            forwardToServer();
        }
        else if ( packetFromServer() )
        {
            if ( serverPolicyAllowsPacket() )
            {
                forwardToClient();
            }
        }
        else
        {
            log( "Unknown packet!" );
        }
    }

    bool controlPacket()
    {
        return _tcpSegment.header().getSyn() || _tcpSegment.header().getFin();
    }

    bool noDataOnPacket()
    {
        return _tcpSegment._payload.size() == 0;
    }

    bool serverPolicyAllowsPacket()
    {
        return _serverPacketPolicy.shouldAllowPacket( _tcpSegment );
    }

    bool clientPolicyAllowsPackets()
    {
        return _clientPacketPolicy.shouldAllowPacket( _packetsToServer );
    }

    bool packetToServer() const
    {
        return extractIp( _ipHeader._mandatoryPart._sourceIp ) != String( ServerConfiguration::SERVER_IP );
    }

    bool packetFromServer() const
    {
        return extractIp( _ipHeader._mandatoryPart._sourceIp ) == String( ServerConfiguration::SERVER_IP );
    }

    // void forwardToServer()
    // {
    //     for ( unsigned i = 0; i < _packetsToServer.size(); ++i )
    //     {
    //         _rawSocket.send( _packetsToServer[i], ServerConfiguration::SERVER_IP );
    //     }

    //     _packetsToServer.clear();
    // }

    void forwardToServer()
    {
        HeapData adjustedPacket;
        _tcpSegment.serialize( adjustedPacket, ProxyConfiguration::PROXY_IP, ServerConfiguration::SERVER_IP );
        _rawSocket.send( adjustedPacket, ServerConfiguration::SERVER_IP );
    }

    void addPacketToQueue()
    {
        HeapData adjustedPacket;
        _tcpSegment.serialize( adjustedPacket, ProxyConfiguration::PROXY_IP, ServerConfiguration::SERVER_IP );
        _packetsToServer.append( adjustedPacket );
    }

    void forwardToClient()
    {
        HeapData adjustedPacket;
        _tcpSegment.serialize( adjustedPacket, ProxyConfiguration::PROXY_IP, ProxyConfiguration::CLIENT_IP );
        _rawSocket.send( adjustedPacket, ProxyConfiguration::CLIENT_IP );
    }

    static String extractIp( unsigned ip )
    {
        struct in_addr ip_addr;
        ip_addr.s_addr = ntohl( ip );
        return inet_ntoa( ip_addr );
    }

    void log( String message ) const
    {
        printf( "ProxyLog: %s\n", message.ascii() );
    }

    void initializeSocket()
    {
        _rawSocket.activate();
    }
};

#endif // __Proxy_h__

//
// Local Variables:
// compile-command: "make -C ../.. "
// tags-file-name: "../../TAGS"
// c-basic-offset: 4
// End:
//
